#include <stdio.h>
#include <math.h>
#include <string.h>  /* String function definitions */
#include <SDL/SDL.h>
#include <SDL/SDL_image.h>

#include <iostream>
using namespace std;

const int NO_SHADES = 2;

int ConvertShade(int v)
{
  return ((int)(v/(256/NO_SHADES)))*256/(NO_SHADES-1)-1;
} // Convert Shade

int GetGray(SDL_Surface *p_source_image, int x, int y)
{
  return ((int)(((unsigned char*)(p_source_image->pixels))[(x*3+y*p_source_image->pitch)+0]) +
	  (int)(((unsigned char*)(p_source_image->pixels))[(x*3+y*p_source_image->pitch)+1]) +
	  (int)(((unsigned char*)(p_source_image->pixels))[(x*3+y*p_source_image->pitch)+2]))/3;
} // GetGray

void SetGray(SDL_Surface *p_source_image, int x, int y, int gray)
{
  if(gray < 0) gray = 0;
  if(gray > 255) gray = 255;

  ((char*)(p_source_image->pixels))[(x*3+y*p_source_image->pitch)+0] = gray;
  ((char*)(p_source_image->pixels))[(x*3+y*p_source_image->pitch)+1] = gray;
  ((char*)(p_source_image->pixels))[(x*3+y*p_source_image->pitch)+2] = gray;
} // GetGray

int AddGray(SDL_Surface *p_source_image, int x, int y, int diff)
{
  int gray = GetGray(p_source_image, x, y);

  gray += diff;
  SetGray(p_source_image, x, y, gray);

} // SetGray

int main(int argc, char** argv)
{
  const int NO_SOURCE_IMAGES = 1;

  SDL_Surface *screen;
  SDL_Surface *p_source_image;
  SDL_Event   event;
  SDL_Rect    rect;
  int         window_width, window_height;
  int         current_x, current_y;
  bool        fullscreen = false;
  Uint32      flags = SDL_HWSURFACE;

  p_source_image=IMG_Load(argv[1]);

  window_width = p_source_image->w; window_height = p_source_image->h;
  
  if(SDL_Init(SDL_INIT_VIDEO) < 0) {
    cerr << "Could not initialize SDL video" << endl;
    exit(-1);
  } // if
  atexit(SDL_Quit);
  
  if(fullscreen) flags |= SDL_FULLSCREEN;
  
  screen = SDL_SetVideoMode(window_width, window_height, 24, flags);
  
  if(!screen) {
    cerr << argv[0] << "could not SetVideoMode() - abort" << endl;
    exit(-1);
  } // if

  for(int y = 1; y < p_source_image->h-1; y++) {
    for(int x = 1; x < p_source_image->w-1; x++) {

      int gray = GetGray(p_source_image, x, y);

      int display_gray = ConvertShade(gray);

      int error = gray - display_gray;
            
      // Step 1: spread 7/16 of error into the pixel to the right, at (x+1, y)
       AddGray(p_source_image, x+1, y, 7*error/16);

      // Step 2: spread 3/16 of error into the pixel to the right, at (x-1, y+1)
         AddGray(p_source_image, x-1, y+1, 3*error/16);

      // Step 3: spread 5/16 of error into the pixel to the right, at (x, y+1)
            AddGray(p_source_image, x, y+1, 5*error/16);

      // Step 4: spread 1/16 of error into the pixel to the right, at (x+1, y+1)
            AddGray(p_source_image, x+1, y+1, 1*error/16);
      
      SetGray(p_source_image, x, y, display_gray);
      // SetGray(p_source_image, x, y, abs(error)); // display the error only
    } // for
  } // for

  if(p_source_image) {
    SDL_BlitSurface(p_source_image, NULL, screen, NULL);
    SDL_UpdateRect(screen, 0, 0, p_source_image->w, p_source_image->h);
  } // if

  for(;;) {
	
    if((SDL_WaitEvent(&event) != 0)) {
      
      switch(event.type) {
	
      case SDL_MOUSEMOTION:
	current_x = event.motion.x;	
	current_y = event.motion.y;
	break;
	
      case SDL_MOUSEBUTTONDOWN:
	break;
	
      case SDL_MOUSEBUTTONUP:
	break;
	
      case SDL_KEYDOWN:
	
	switch(event.key.keysym.sym) {

	case 'w':
	  break;

	case 'l':
	  if(p_source_image) {
	    SDL_BlitSurface(p_source_image, NULL, screen, NULL);
	    SDL_UpdateRect(screen, 0, 0, p_source_image->w, p_source_image->h);
	  } // if
	  break;

	case 27:
	case 'q':
	case 'Q':
	  goto exit_app;
			    
	default:
	  cout << "key: " << (int)event.key.keysym.sym << endl;
	  break;
	} // switch
      } // switch
    } // if
    
    //    SDL_UnlockSurface(screen);
    //    SDL_UpdateRect(screen, 0, 0, window_width, window_height);	  
  } // for
    
 exit_app:
  
  //  SDL_UnlockSurface(p_global_image);
  //  SDL_FreeSurface(p_global_image);

  SDL_UnlockSurface(p_source_image);
  SDL_FreeSurface(p_source_image);

  SDL_UnlockSurface(screen);	    
  SDL_FreeSurface(screen);
} // main
